type quadtree =
  | Base of int * int array array
  | Node of int * quadnode
  and quadnode = quadtree array;;

(* Le type Pool de Gasche. *)
module type POOL = sig
type pool
val create: unit -> pool
val get: pool -> int -> quadtree -> quadtree option
val set: pool -> int -> quadtree -> quadtree -> unit
end;;

module AMap = Map.Make (struct type t = quadtree let compare = compare end);;

module Pool: POOL = struct
type pool = quadtree AMap.t array
let pool_size = 4194301
let create () = Array.make pool_size AMap.empty
let get p d arb =
try Some (AMap.find arb p.(d)) with Not_found -> None
let set p d arb next =
p.(d) <- AMap.add arb next p.(d)
end;;

let global_pool = Pool.create ();;

(*
NO           NE
n11  n12     n13  n14
n21  n22     n23  n24
SO           SE 
n31  n32     n33  n34
n41  n42     n43  n44
*)

(* Determination des cas de base. *)

type dir = NO | NE | SO | SE;;

let test = 0b1000100010001000;;

let masque = function
    | NO -> 0b1110101011001000
    | NE -> 0b0101110101001100
    | SO -> 0b0011001010111010
    | SE -> 0b0001001101010111;;

let etat dir sq =
  let n = match dir with
      | NO -> 12 | NE -> 9
      | SO -> 6 | SE -> 3
  in (sq lsr n) land 1;;

let compter =
  let rec auxi k = function
      | 0 -> k
      | i -> auxi (k + 1) (i land (i - 1))
  in auxi 0;;

let voisin dir sq = compter (sq land masque dir);;

let suivant dir sq =
  let v = voisin dir sq and e = etat dir sq in
    if (e = 0 && v = 3) || (e = 1 && (v = 2 || v = 3))
    then 1 else 0;;

let calcul sq = Base (1, [|[|suivant NO sq; suivant NE sq|]; [|suivant SO sq; suivant SE sq|]|]);;

let base () = Array.init 65536 calcul;;

let table = base ();;

(* Fonctions de manipulation des arbres. *)

let rec numerote = function
    | Base (d, [|[|no; ne|]; [|so; se|]|]) -> (no lsl 3) lor (ne lsl 2) lor (so lsl 1) lor se
    | Node (d, [|no; ne; so; se|]) -> let p = 1 lsl (2 * (d - 1)) and q = 1 lsl (2 * d - 1) in
          ((numerote no) lsl (p lor q)) lor ((numerote ne) lsl q) lor ((numerote so) lsl p) lor (numerote se)
    | _ -> failwith "Numerote";;

let horizontal a1 a2 = match (a1, a2) with
    | (Node (da, a), Node (db, b)) -> Node (da, [|a.(1); b.(0); a.(3); b.(2)|])
    | _ -> failwith "Horizontal";;

let vertical a1 a2 = match (a1, a2) with
    | (Node (da, a), Node (db, b)) -> Node (da, [|a.(2); a.(3); b.(0); b.(1)|])
    | _ -> failwith "Vertical";;

let centre = function
    | Node (d, [|Base (d0, n0); Base (d1, n1); Base (d2, n2); Base (d3, n3)|]) -> Base (d - 1, [|[|n0.(1).(1); n1.(1).(0)|]; [|n2.(0).(1); n3.(0).(0)|]|])
    | Node (d, [|Node (d0, n0); Node (d1, n1); Node (d2, n2); Node (d3, n3)|]) -> Node (d - 1, [|n0.(3); n1.(2); n2.(1); n3.(0)|])
    | _ -> failwith "Centre";;

let rec traite arb = match arb with
    | Base (d, a) -> failwith "Traite"
    | Node (2, a) -> table.(numerote arb)
    | Node (d, a) ->
        let a00 = centre a.(0)
        and a01 = centre (horizontal a.(0) a.(1))
        and a02 = centre a.(1)
        and a10 = centre (vertical a.(0) a.(2))
        and a11 = centre (centre arb)
        and a12 = centre (vertical a.(1) a.(3))
        and a20 = centre a.(2)
        and a21 = centre (horizontal a.(2) a.(3))
        and a22 = centre a.(3) in
          let ano = Node (d - 1, [|a00; a01; a10; a11|])
          and ane = Node (d - 1, [|a01; a02; a11; a12|])
          and aso = Node (d - 1, [|a10; a11; a20; a21|])
          and ase = Node (d - 1, [|a11; a12; a21; a22|]) in
            Node (d, [|(h (d - 1) ano); (h (d - 1) ane); (h (d - 1) aso); (h (d - 1) ase)|]);
and h d arb =
  match Pool.get global_pool d arb with
    | Some result -> result
    | None ->
        let a = traite arb in
          Pool.set global_pool d arb a;
          a;;

(* Generations. *)

let generation = function
    | Node (d, [|Node (dno, ano); Node (dne, ane); Node (dso, aso); Node (dse, ase)|]) ->
        let a00 = Node (d - 1, [|ase.(3); aso.(2); ane.(1); ano.(0)|])
        and a01 = Node (d - 1, [|aso.(3); ase.(2); ano.(1); ane.(0)|])
        and a10 = Node (d - 1, [|ane.(3); ano.(2); ase.(1); aso.(0)|])
        and a11 = Node (d - 1, [|ano.(3); ane.(2); aso.(1); ase.(0)|]) in
          let a0 = Node (d, [|a00; a01; a10; a11|])
          and a1 = Node (d, [|a01; a00; a11; a10|])
          and a2 = Node (d, [|a10; a11; a00; a01|])
          and a3 = Node (d, [|a11; a10; a01; a00|]) in
            Node (d, [|h d a0; h d a1; h d a2; h d a3|])
    | _ -> failwith "Generation";;

let rec multigen_nbr arb = function
    | 0 -> print_int 0; arb
    | n -> print_int n; print_newline (); multigen_nbr (generation arb) (n - 1);;

(* Essais. *)
let rec random = function
    | 0 -> failwith "Random"
    | 1 -> let x = Random.int 9 in Base (1, [|[|x lsr 3 land 1; x lsr 2 land 1|]; [|x lsr 1 land 1; x land 1|]|])
    | n -> let q = [|random (n - 1); random (n - 1); random (n - 1); random (n - 1)|] in Node (n, q);;

multigen_nbr (random 8) 100;;